"""The Modified Differential Multiplier Method (MDMM) for PyTorch."""

from .mdmm import (ConstraintReturn, Constraint, EqConstraint, MaxConstraint, MaxConstraintHard,
                   MinConstraint, MinConstraintHard, BoundConstraintHard, MDMMReturn, MDMM)

__version__ = '0.1.1'



#!/usr/bin/env python3

"""An MNIST classifier using MDMM to enforce layer norm constraints."""

import argparse
import csv
from functools import partial

import torch
from torch import nn
from torch.utils import data
from torchvision import datasets, transforms

import mdmm

BATCH_SIZE = 50
EPOCHS = 100


def main():
    p = argparse.ArgumentParser(formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    p.add_argument('--norm', type=float, default=1.,
                   help='the target layer norm')
    p.add_argument('--scale', type=float, default=1.,
                   help='the infeasibility scale factor')
    p.add_argument('--damping', type=float, default=20.,
                   help='the damping strength')
    p.add_argument('--lr', type=float, default=0.02,
                   help='the learning rate')
    args = p.parse_args()

    device = torch.device('cuda' if torch.cuda.is_available() else 'cpu')
    print('Using device:', device)
    torch.manual_seed(0)

    tf = transforms.ToTensor()
    train_set = datasets.MNIST('data/mnist', download=True, transform=tf)
    train_dl = data.DataLoader(train_set, BATCH_SIZE, shuffle=True,
                               num_workers=2, pin_memory=True)
    val_set = datasets.MNIST('data/mnist', train=False, download=True, transform=tf)
    val_dl = data.DataLoader(train_set, BATCH_SIZE, num_workers=2, pin_memory=True)

    model = nn.Sequential(
        nn.Conv2d(1, 16, 3, padding=1),
        nn.ReLU(inplace=True),
        nn.MaxPool2d(2),
        nn.Conv2d(16, 16, 3, padding=1),
        nn.ReLU(inplace=True),
        nn.MaxPool2d(2),
        nn.Flatten(),
        nn.Linear(16 * 7 * 7, 10),
    ).to(device)
    print('Parameters:', sum(param.numel() for param in model.parameters()))

    crit = nn.CrossEntropyLoss()

    constraints = []
    for layer in model:
        if hasattr(layer, 'weight'):
            fn = partial(lambda x: x.weight.abs().mean(), layer)
            constraints.append(mdmm.EqConstraint(fn, args.norm,
                                                 scale=args.scale, damping=args.damping))

    mdmm_module = mdmm.MDMM(constraints)
    opt = mdmm_module.make_optimizer(model.parameters(), lr=args.lr)

    writer = csv.writer(open('mdmm_demo_mnist.csv', 'w'))
    writer.writerow(['loss', 'norm_1', 'norm_2', 'norm_3'])

    def train():
        model.train()
        i = 0
        losses = []
        for inputs, targets in train_dl:
            i += 1
            inputs = inputs.to(device, non_blocking=True)
            targets = targets.to(device, non_blocking=True)
            outputs = model(inputs)
            loss = crit(outputs, targets)
            losses.append(loss)
            mdmm_return = mdmm_module(loss)
            writer.writerow([loss.item(), *(norm.item() for norm in mdmm_return.fn_values)])
            opt.zero_grad()
            mdmm_return.value.backward()
            opt.step()
            if i % 100 == 0:
                print(f'{i} {sum(losses[-100:]) / 100:g}')
                print('Layer weight norms:',
                      *(f'{norm.item():g}' for norm in mdmm_return.fn_values))

    def val():
        print('Validating...')
        model.eval()
        losses = []
        for inputs, targets in val_dl:
            inputs = inputs.to(device, non_blocking=True)
            targets = targets.to(device, non_blocking=True)
            outputs = model(inputs)
            loss = crit(outputs, targets)
            losses.append(loss * len(outputs))
        loss = sum(losses) / len(val_set)
        print(f'Validation loss: {loss.item():g}')

    try:
        for epoch in range(1, EPOCHS + 1):
            print('Epoch', epoch)
            train()
            val()
    except KeyboardInterrupt:
        pass


if __name__ == '__main__':
    main()



#!/usr/bin/env python3

"""Total variation image denoising using MDMM to enforce a max constraint on
the image total variation."""

import argparse
import csv

from PIL import Image
import torch
from torch import nn
from torch.nn import functional as F
from torchvision.transforms import functional as TF

import mdmm


class TVLoss(nn.Module):
    def forward(self, input):
        input = F.pad(input, (0, 1, 0, 1), 'replicate')
        x_diff = input[..., :-1, 1:] - input[..., :-1, :-1]
        y_diff = input[..., 1:, :-1] - input[..., :-1, :-1]
        diff = x_diff**2 + y_diff**2 + 1e-8
        return diff.sum(dim=1).sqrt().sum()


def main():
    p = argparse.ArgumentParser(formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    p.add_argument('input_image', type=str,
                   help='the input image')
    p.add_argument('output_image', type=str, nargs='?', default='out.png',
                   help='the output image')
    p.add_argument('--max-tv', type=float, default=0.02,
                   help='the maximum allowable total variation per sample')
    p.add_argument('--scale', type=float, default=1.,
                   help='the infeasibility scale factor')
    p.add_argument('--damping', type=float, default=1e-2,
                   help='the damping strength')
    p.add_argument('--lr', type=float, default=2e-3,
                   help='the learning rate')
    args = p.parse_args()

    device = torch.device('cuda' if torch.cuda.is_available() else 'cpu')
    print('Using device:', device)

    pil_image = Image.open(args.input_image).resize((128, 128), Image.LANCZOS)
    target = TF.to_tensor(pil_image)[None].to(device)
    input = target.clone().requires_grad_()
    # torch.manual_seed(0)
    # target += torch.randn_like(target) / 10
    # target.clamp_(0, 1)

    crit_l2 = nn.MSELoss(reduction='sum')
    crit_tv = TVLoss()
    max_tv = args.max_tv * input.numel()

    constraint = mdmm.MaxConstraint(lambda: crit_tv(input), max_tv, args.scale, args.damping)
    mdmm_module = mdmm.MDMM([constraint])
    opt = mdmm_module.make_optimizer([input], lr=args.lr)

    writer = csv.writer(open('mdmm_demo_tv.csv', 'w'))
    writer.writerow(['l2_loss', 'tv_loss'])

    try:
        i = 0
        while True:
            i += 1
            loss = crit_l2(input, target)
            mdmm_return = mdmm_module(loss)
            writer.writerow([loss.item() / input.numel(),
                             mdmm_return.fn_values[0].item() / input.numel()])
            msg = '{} l2={:g}, tv={:g}'
            print(msg.format(i, loss.item() / input.numel(),
                             mdmm_return.fn_values[0].item() / input.numel()))
            if not mdmm_return.value.isfinite():
                break
            opt.zero_grad()
            mdmm_return.value.backward()
            opt.step()
    except KeyboardInterrupt:
        pass

    TF.to_pil_image(input[0].clamp(0, 1)).save(args.output_image)


if __name__ == '__main__':
    main()



"""The Modified Differential Multiplier Method (MDMM) for PyTorch."""

import abc
from dataclasses import dataclass
from typing import List

import torch
from torch import nn, optim


@dataclass
class ConstraintReturn:
    """The return type for constraints."""
    value: torch.Tensor
    fn_value: torch.Tensor
    inf: torch.Tensor


class Constraint(nn.Module, metaclass=abc.ABCMeta):
    """The base class for all constraint types."""

    def __init__(self, fn, scale, damping):
        super().__init__()
        self.fn = fn
        self.register_buffer('scale', torch.as_tensor(scale))
        self.register_buffer('damping', torch.as_tensor(damping))
        self.lmbda = nn.Parameter(torch.tensor(0.))

    def extra_repr(self):
        return f'scale={self.scale:g}, damping={self.damping:g}'

    @abc.abstractmethod
    def infeasibility(self, fn_value):
        ...

    def forward(self):
        fn_value = self.fn()
        inf = self.infeasibility(fn_value)
        l_term = self.lmbda * inf
        damp_term = self.damping * inf**2 / 2
        return ConstraintReturn(self.scale * (l_term + damp_term), fn_value, inf)


class EqConstraint(Constraint):
    """Represents an equality constraint."""

    def __init__(self, fn, value, scale=1., damping=1.):
        super().__init__(fn, scale, damping)
        self.register_buffer('value', torch.as_tensor(value))

    def extra_repr(self):
        return f'value={self.value:g}, scale={self.scale:g}, damping={self.damping:g}'

    def infeasibility(self, fn_value):
        return self.value - fn_value


class MaxConstraint(Constraint):
    """Represents a maximum inequality constraint which uses a slack variable."""

    def __init__(self, fn, max, scale=1., damping=1.):
        super().__init__(fn, scale, damping)
        self.register_buffer('max', torch.as_tensor(max))
        self.slack = nn.Parameter(torch.as_tensor(float('nan')))

    def extra_repr(self):
        return f'max={self.max:g}, scale={self.scale:g}, damping={self.damping:g}'

    def infeasibility(self, fn_value):
        if self.slack.isnan():
            self.slack = nn.Parameter((self.max - fn_value).relu().pow(1/2))
        return self.max - fn_value - self.slack**2


class MaxConstraintHard(Constraint):
    """Represents a maximum inequality constraint without a slack variable."""

    def __init__(self, fn, max, scale=1., damping=1.):
        super().__init__(fn, scale, damping)
        self.register_buffer('max', torch.as_tensor(max))

    def extra_repr(self):
        return f'max={self.max:g}, scale={self.scale:g}, damping={self.damping:g}'

    def infeasibility(self, fn_value):
        return fn_value.clamp(max=self.max) - fn_value


class MinConstraint(Constraint):
    """Represents a minimum inequality constraint which uses a slack variable."""

    def __init__(self, fn, min, scale=1., damping=1.):
        super().__init__(fn, scale, damping)
        self.register_buffer('min', torch.as_tensor(min))
        self.slack = nn.Parameter(torch.as_tensor(float('nan')))

    def extra_repr(self):
        return f'min={self.min:g}, scale={self.scale:g}, damping={self.damping:g}'

    def infeasibility(self, fn_value):
        if self.slack.isnan():
            self.slack = nn.Parameter((fn_value - self.min).relu().pow(1/2))
        return fn_value - self.min - self.slack**2


class MinConstraintHard(Constraint):
    """Represents a minimum inequality constraint without a slack variable."""

    def __init__(self, fn, min, scale=1., damping=1.):
        super().__init__(fn, scale, damping)
        self.register_buffer('min', torch.as_tensor(min))

    def extra_repr(self):
        return f'min={self.min:g}, scale={self.scale:g}, damping={self.damping:g}'

    def infeasibility(self, fn_value):
        return fn_value.clamp(min=self.min) - fn_value


class BoundConstraintHard(Constraint):
    """Represents a bound constraint."""

    def __init__(self, fn, min, max, scale=1., damping=1.):
        super().__init__(fn, scale, damping)
        self.register_buffer('min', torch.as_tensor(min))
        self.register_buffer('max', torch.as_tensor(max))

    def extra_repr(self):
        return f'min={self.min:g}, max={self.max:g}, ' \
               f'scale={self.scale:g}, damping={self.damping:g}'

    def infeasibility(self, fn_value):
        return fn_value.clamp(self.min, self.max) - fn_value


@dataclass
class MDMMReturn:
    """The return type for MDMM."""
    value: torch.Tensor
    fn_values: List[torch.Tensor]
    infs: List[torch.Tensor]


class MDMM(nn.ModuleList):
    """The main MDMM class, which combines multiple constraints."""

    def make_optimizer(self, params, *, optimizer=optim.Adamax, lr=2e-3):
        lambdas = [c.lmbda for c in self]
        slacks = [c.slack for c in self if hasattr(c, 'slack')]
        return optimizer([{'params': params, 'lr': lr},
                          {'params': lambdas, 'lr': -lr},
                          {'params': slacks, 'lr': lr}])

    def forward(self, loss):
        value = loss.clone()
        fn_values, infs = [], []
        for c in self:
            c_return = c()
            value += c_return.value
            fn_values.append(c_return.fn_value)
            infs.append(c_return.inf)
        return MDMMReturn(value, fn_values, infs)



from pathlib import Path

import setuptools

import mdmm


setuptools.setup(
    name='mdmm',
    version=mdmm.__version__,
    description='The Modified Differential Multiplier Method (MDMM) for PyTorch.',
    long_description=(Path(__file__).resolve().parent / 'README.md').read_text(),
    long_description_content_type='text/markdown',
    url='https://github.com/crowsonkb/mdmm',
    author='Katherine Crowson',
    author_email='crowsonkb@gmail.com',
    license='MIT',
    license_files=['LICENSE'],
    packages=['mdmm'],
    install_requires=['dataclasses>=0.8;python_version<"3.7"',
                      'torch>=1.7.1'],
    python_requires=">=3.6",
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Topic :: Scientific/Engineering :: Artificial Intelligence',
        'Topic :: Scientific/Engineering :: Mathematics',
    ],
)
